/*
 *
 *  Computer as Handsfree device
 *
 *  Copyright (C) 2004  Simon Vogl <vogl@soft.uni-linz.ac.at>
 *  Copyright (C) 2009  Stjepan Gros <sgros@zemris.fer.hr>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation;
 *
 *  Software distributed under the License is distributed on an "AS
 *  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 *  implied. See the License for the specific language governing
 *  rights and limitations under the License.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <termios.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <time.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sco.h>
#include <bluetooth/rfcomm.h>

#include <unistd.h>

#include <glib.h>

#include "pulse.h"
#include "btcommon.h"

#define RFCOMM_BUFFER_SIZE		2048
#define SCO_BUFFER_SIZE			2048
#define MIC_BUFFER_SIZE			48

struct rfcomm_connection {

	/*
	 * RFCOMM specific data
	 * TODO: This has to be extended since we can be HF for multiple mobile phones
	 */
	int rfcomm;
	GIOChannel *ioc_rfcomm;
};

struct appdata;

struct sco_connection {
	struct sockaddr_sco addr;
	struct sco_conninfo conn;
	struct sco_options opts;

	/*
	 * Handles for audio chanel from/to mobile phone
	 * TODO: This has to be extended since we can be HF for multiple mobile phones
	 */
	int sco;
	GIOChannel *ioc_sco;

	struct appdata *appdata;
};

struct sco_connection *sco_connection_new(void)
{
	return g_malloc0(sizeof(struct sco_connection));
}

void sco_connection_free(struct sco_connection *sc)
{
	g_free(sc);
}

struct appdata {
	gint refcnt;

	GMainLoop *mainloop;

	bdaddr_t bdaddr;
	bdaddr_t bdaddr_local;

	/*
	 * RFCOMM specific data
	 * TODO: This has to be extended since we can be HF for multiple mobile phones
	 */
	struct rfcomm_connection rc;

	/*
	 * Handles for audio chanel from/to mobile phone
	 */
	int sco_listen;
	GIOChannel *ioc_sco_listen;

	/*
	 * SCO connection specific data
	 * TODO: This has to be extended since we can be HF for multiple mobile phones
	 */
	struct sco_connection *sc;

	GMutex *sc_lock;

	GThread *mic;

	/*
	 * File descriptor where data received from the mobnile device will be saved
	 */
	int mobilefd;
};

/**
 * Configuration variables initialized from the command line
 */
static gchar *opt_bdaddr = NULL;
static gint opt_channel = 1;
static gboolean opt_nomic = FALSE;
static gchar *opt_mobiledata = NULL;

/*
 * Command line options
 */
static GOptionEntry entries[] = 
{
  { "bdaddr",  'b', 0, G_OPTION_ARG_STRING, &opt_bdaddr, "BT address of a mobile device", "aa:bb:cc:dd:ee:ff" },
  { "channel", 'c', 0, G_OPTION_ARG_INT, &opt_channel, "Channel to communicate with mobile device", "N" },
  { "no-mic", 'm', 0, G_OPTION_ARG_NONE, &opt_nomic, "Do not turn on microphone", NULL },
  { "write-from-mobile", 'd', 0, G_OPTION_ARG_STRING, &opt_mobiledata, "Write data from mobile into file", "<filename>" },
  { NULL }
};

gboolean rfcomm_process_cb(GIOChannel *source, GIOCondition condition, gpointer data)
{
	struct appdata *appdata = (struct appdata *)data;
	char buffer[RFCOMM_BUFFER_SIZE];
	ssize_t bytes;

	// fprintf (stderr, "rfcomm_process_cb (condition = %d)\n", condition);

	if (condition & G_IO_IN) {
		if ((bytes = read(appdata->rc.rfcomm, buffer, RFCOMM_BUFFER_SIZE)) < 0) {
			perror("read");
			return TRUE;
		}
		// fprintf (stderr, "\tbytes = %zd\n%s\n", bytes, buffer);
	}

	if (condition & G_IO_HUP || condition & G_IO_ERR || condition & G_IO_NVAL) {
		/*
		 * Reopen RFCOMM connection!
		 */
		close(appdata->rc.rfcomm);
		g_io_channel_unref(appdata->rc.ioc_rfcomm);

		if ((appdata->rc.rfcomm = rfcomm_connect(&appdata->bdaddr_local, &appdata->bdaddr, opt_channel)) < 0) {
			/*
			 * Free data structure!
			 */
			return FALSE;
		}

		appdata->rc.ioc_rfcomm = g_io_channel_unix_new(appdata->rc.rfcomm);
		g_io_add_watch (appdata->rc.ioc_rfcomm, G_IO_IN | G_IO_HUP, rfcomm_process_cb, &appdata);
	}

	return TRUE;
}

gboolean sco_process_cb(GIOChannel *source, GIOCondition condition, gpointer data)
{
	struct sco_connection *sc = (struct sco_connection *)data;
	struct appdata *appdata = sc->appdata;
	char buffer[SCO_BUFFER_SIZE];
	ssize_t bytes;

	// fprintf (stderr, "sco_process_cb (condition = %d)\n", condition);

	if (condition & G_IO_IN) {
		if ((bytes = read(sc->sco, buffer, SCO_BUFFER_SIZE)) > 0) {
			sound_play(buffer, bytes);

#if 0
			if (write(STDOUT_FILENO, buffer, bytes) < 0) {
				fprintf(stderr, "write failed\n");
				perror("write");
			}
#endif

			if (appdata->mobilefd)
				if (write(appdata->mobilefd, buffer, bytes) < 0) {
					fprintf(stderr, "mobilefd = %d\n", appdata->mobilefd);
					perror("write");
				}
		}
	}

	if (condition & G_IO_HUP || condition & G_IO_ERR || condition & G_IO_NVAL) {
		/*
		 * Free SC structure!!!
		 */
		sc->appdata = NULL;
		g_atomic_int_dec_and_test(&appdata->refcnt);
		
		g_mutex_lock(appdata->sc_lock);
		g_io_channel_unref(appdata->sc->ioc_sco);
		appdata->sc = NULL;
		g_mutex_unlock(appdata->sc_lock);
		fprintf(stderr, "Removed SCO channel\n");
		return FALSE;
	}

	return TRUE;
}

gboolean sco_listen_process_cb(GIOChannel *source, GIOCondition condition, gpointer data)
{
	struct appdata *appdata = (struct appdata *)data;
	struct sco_connection *sc;

	// fprintf (stderr, "sco_listen_process_cb (condition = %d)\n", condition);

	sc = sco_connection_new();

	if ((sc->sco = sco_socket_accept(appdata->sco_listen, NULL, NULL, NULL, NULL)) < 0) {
		sco_connection_free(sc);
		return FALSE;
	}

	sc->ioc_sco = g_io_channel_unix_new(sc->sco);
	g_io_add_watch (sc->ioc_sco, G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL, sco_process_cb, sc);

	g_atomic_int_inc(&appdata->refcnt);
	sc->appdata = appdata;

	g_mutex_lock(appdata->sc_lock);
	appdata->sc = sc;
	g_mutex_unlock(appdata->sc_lock);
	fprintf (stderr, "New SCO channel connected\n");

	return TRUE;
}

gpointer mic_thread(gpointer data)
{
	struct appdata *appdata = (struct appdata *)data;
	char buffer[MIC_BUFFER_SIZE];
	int s, c = 0, c1 = 0;

	fprintf (stderr, "Microphone thread started\n");

	while (TRUE) {

		if (g_mutex_trylock(appdata->sc_lock)) {

			if (!appdata->sc) {
				g_mutex_unlock(appdata->sc_lock);
				g_usleep(3000);
				continue;
			}

			g_mutex_unlock(appdata->sc_lock);

			if (++c > 1000) {
				fprintf (stderr, "Reading data...\n");
				c = 0;
			}

			if (sound_rec(buffer, MIC_BUFFER_SIZE) < 0)
				break;

			if (g_mutex_trylock(appdata->sc_lock)) {
				s = MIC_BUFFER_SIZE;
				if (appdata->sc) {
					if (++c1 > 1000) {
						fprintf (stderr, "Writing data...\n");
						c1 = 0;
					}
					s = write(appdata->sc->sco, buffer, MIC_BUFFER_SIZE);
				}

				g_mutex_unlock(appdata->sc_lock);

				if (s != MIC_BUFFER_SIZE)
					break;
			}

		} else
			g_usleep(3000);
	}

	return NULL;
}

struct appdata appdata;

static void sig_term(int sig)
{
	g_main_loop_quit(appdata.mainloop);
}

int main(int argc, char **argv)
{
	GError *error = NULL;
	GOptionContext *context;
	struct sigaction sa;

	memset(&appdata, 0, sizeof(struct appdata));
	appdata.refcnt = 1;

	g_thread_init(NULL);

	/*
	 * Parse command line options
	 */
	context = g_option_context_new ("- turn your PC into a handsfree device");
	g_option_context_add_main_entries (context, entries, NULL);
	if (!g_option_context_parse (context, &argc, &argv, &error)) {
		g_print ("option parsing failed: %s\n", error->message);
		exit (1);
	}

	if (opt_bdaddr == NULL) {
		g_print ("Please specify BT address of a mobile device\n");
		exit (1);
	}

#if 0
	/*
	 * Create file to store data received from mobile device
	 * This isn't working so it's disabled!
	 */
	if (opt_mobiledata) {
		if ((appdata.mobilefd = open(opt_mobiledata, O_CREAT, S_IRWXU)) < 0) {
			perror("open");
			exit (1);
		}
		fprintf (stderr, "mobilefd init = %d\n", appdata.mobilefd);
	}
#endif

	/*
	 * Obtain local BT address, convert remote address and connect
	 * RFCOMM channel to mobile device
	 */
	hci_devba(0, &appdata.bdaddr_local);
	str2ba(opt_bdaddr, &appdata.bdaddr);
	if ((appdata.rc.rfcomm = rfcomm_connect(&appdata.bdaddr_local, &appdata.bdaddr, opt_channel)) < 0)
		exit(1);

	appdata.rc.ioc_rfcomm = g_io_channel_unix_new(appdata.rc.rfcomm);
	g_io_add_watch (appdata.rc.ioc_rfcomm, G_IO_IN | G_IO_HUP, rfcomm_process_cb, &appdata);

	appdata.sc_lock = g_mutex_new();
	g_mutex_lock(appdata.sc_lock);
	appdata.sc = NULL;
	g_mutex_unlock(appdata.sc_lock);

	/*
	 * Create SCO listener for incoming requests from mobile phone
	 */
	if ((appdata.sco_listen = sco_socket_listen(&appdata.bdaddr_local)) < 0)
		exit(1);

	appdata.ioc_sco_listen = g_io_channel_unix_new(appdata.sco_listen);
	g_io_add_watch (appdata.ioc_sco_listen, G_IO_IN | G_IO_HUP, sco_listen_process_cb, &appdata);

	/*
	 * Initialize sound device
	 */
	if (sound_init(argv[0]) < 0)
		return -1;

	/*
	 * Create thread that reads data from microphone and writes it to a phone
	 */
	if (!opt_nomic)
		appdata.mic = g_thread_create(mic_thread, &appdata, TRUE, NULL);

	/*
	 * Handle signals...
	 */
	memset(&sa, 0, sizeof(sa));
	sa.sa_flags = SA_NOCLDSTOP;
	sa.sa_handler = sig_term;
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT,  &sa, NULL);

	sa.sa_handler = SIG_IGN;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);

	/*
	 * Initialize GLib's main loop
	 */
	appdata.mainloop = g_main_loop_new(NULL, FALSE);
	g_main_loop_run(appdata.mainloop);

	if (appdata.mobilefd)
		close(appdata.mobilefd);

	return 0;
}
