/*
 *
 *  Computer as Headset utility
 *
 *  Copyright (C) 2004  Simon Vogl <vogl@soft.uni-linz.ac.at>
 *  Copyright (C) 2009  Stjepan Gros <sgros@zemris.fer.hr>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation;
 *
 *  Software distributed under the License is distributed on an "AS
 *  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 *  implied. See the License for the specific language governing
 *  rights and limitations under the License.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <termios.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sco.h>
#include <bluetooth/rfcomm.h>

#include <unistd.h>
#include <stdarg.h>     /* we need va_list */
#include <errno.h>

#include "pulse.h"
#include "btcommon.h"


static volatile int terminate = 0;

static void sig_term(int sig) {
	terminate = 1;
}

/******************* RFCOMM utilities
 *******************/



static void usage(char *appname)
{
	printf("Usage: %s <bdaddr> [channel]\n", appname);
}

#define PLAY	1
#define RECORD	2

static int writecmd(int rd, char*cmd)
{
    fprintf(stderr,"sending.cmd %s\n",cmd);
    return write(rd, cmd,strlen(cmd));
}

char retbuf[255];
int retCode;

#define RETC_NONE 0 
#define RETC_OK 1 
#define RETC_ERR 2 
#define RETC_RING 3 
#define RETC_RSLT 4  // +blah:
#define RETC_TEXT 4 

#define RET_OK 0
#define RET_ERR -1
#define RET_TIMEOUT -2

int main(int argc, char *argv[])
{
    struct sigaction sa;

    fd_set rfds;
    struct timeval timeout;
    char buf[2048];
    int maxfd, sel, rlen, wlen, fdout;

    bdaddr_t local;
    bdaddr_t bdaddr;
    uint8_t channel;

    char *filename;
    mode_t filemode;

    int dd, rd, sd, fd, scolink=0;
    uint16_t sco_handle, sco_mtu, vs;

	int state=0;

	switch (argc) {
	case 2:
		str2ba(argv[1], &bdaddr);
		channel = 1;
		break;
	case 3:
		str2ba(argv[1], &bdaddr);
		channel = atoi(argv[2]);
		break;
	default:
		usage(argv[0]);
		return -1;
	}

	if (sound_init(argv[0]) < 0)
		return -1;

    filemode = O_RDONLY;
    filename = argv[1];
    
    /** set up bt stack */
    hci_devba(0, &local);
    dd = hci_open_dev(0);
    hci_read_voice_setting(dd, &vs, 1000);
    vs = htobs(vs);
    fprintf(stderr, "Voice setting: 0x%04x\n", vs);fflush(stderr);
    close(dd);
	
    if (vs != 0x0060) {
	fprintf(stderr, "The voice setting must be 0x0060\n");fflush(stderr);
	return -1;
    }
  
    fd = -1;

    fdout = open("out.raw",O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
    if (fdout<0) { perror("could not save to output!"); };

    memset(&sa, 0, sizeof(sa));
    sa.sa_flags = SA_NOCLDSTOP;
    sa.sa_handler = sig_term;
    sigaction(SIGTERM, &sa, NULL);
    sigaction(SIGINT,  &sa, NULL);

    sa.sa_handler = SIG_IGN;
    sigaction(SIGCHLD, &sa, NULL);
    sigaction(SIGPIPE, &sa, NULL);

	if ((rd = rfcomm_connect(&local, &bdaddr, channel)) < 0) {
		perror("Can't connect RFCOMM channel");
		return -1;
	}

	fprintf(stderr, "RFCOMM channel connected\n");


#if 0 /* SCO connection is set up by the audio gateway */
    fprintf(stderr, "Setting up SCO channel connected\n");
    if ((sd = sco_connect(&local, &bdaddr, &sco_handle, &sco_mtu)) < 0) {
	perror("Can't connect SCO audio channel");
	close(rd);
	return -1;
    }
    fprintf(stderr, "SCO audio channel connected (handle %d, mtu %d)\n", sco_handle, sco_mtu);
#endif

	if ((sd = sco_socket_listen(&local)) < 0) {
		fprintf(stderr, "Can't init listening SCO audio channel");
		close(rd);
		return -1;
	}

    //wlen = writecmd(rd, "AT+CKPD=200\r"); // request audio link

	maxfd = (rd > sd) ? rd : sd;

	scolink=0;
	state=0;

    while (!terminate) {

	FD_ZERO(&rfds);
	FD_SET(0, &rfds);
	FD_SET(rd, &rfds);
	FD_SET(sd, &rfds);

	if (scolink>0) {
	    FD_SET(scolink, &rfds);
	}

	timeout.tv_sec = 0;
	timeout.tv_usec = 10000;

	if ((sel = select(maxfd + 1, &rfds, NULL, NULL, &timeout)) > 0) {
	    fprintf(stderr,".");fflush(stderr);


	    if (scolink>0 && FD_ISSET(scolink, &rfds) ) {
		fprintf(stderr,":");fflush(stderr);

		rlen = read(scolink, buf, sizeof(buf));
		if (rlen > 0) {
		    fprintf(stderr,"!");fflush(stderr);

		    sound_play(buf,rlen);

		    if (fdout > 0) {
			wlen = write(fdout, buf, rlen);
		    }
		    if (fd>0) {
			int r;
			r=read(fd, buf, rlen);
			write(scolink, buf, rlen);
			if (r<0) {
			    close(fd); 
			    fd=0;
			}
		    } else {
			write(scolink, buf, rlen);
		    }
		} else if (rlen<0) {
		    fprintf(stderr,"lost sco audio connection!\n");
		    close(scolink);
		    scolink=0;
		    maxfd = (rd > sd) ? rd : sd;
		}
	    }

	    if (FD_ISSET(rd, &rfds)) {
		memset(buf, 0, sizeof(buf));
		//rlen = read(rd, buf, sizeof(buf));
		rlen = recv(rd, buf, sizeof(buf),0);
		if (rlen > 0) {
		    fprintf(stderr, "%s\n", buf);
		    if (strstr(buf,"ERROR")!=NULL) {
			fprintf(stderr, "Error\n");
			
		    }else if (strstr(buf,"OK")!=NULL) {
			fprintf(stderr, "oking\n");
		    } else if (strstr(buf,"RING")) {
			fprintf(stderr, "taking call\n");
			wlen = writecmd(rd, "AT+CKPD=200\r"); // the hs doesn't answer
			fprintf(stderr,"wrote %d bytes\n",wlen);
			//wlen = writecmd(rd, "ATA\r\n"); // the hs doesn't answer

		    }
		} else if (rlen<0) {
		    fprintf(stderr,"RFCOMM link error %d (0x%x)!\n",errno,errno);
		    //terminate=1;
		    if (errno == ECONNRESET) {
			close(rd);
			if ((rd = rfcomm_connect(&local, &bdaddr, channel)) < 0) {
			    fprintf(stderr,"RFCOMM link error. terminating. %d (0x%x)!\n",errno,errno);
			    terminate=1;
			}
		    }
		}
	    }

	    if (FD_ISSET(0, &rfds)) { //stdin
		rlen = read(0, buf, sizeof(buf));
		//wlen = writecmd(rd, "AT+CKPD=200\r\n"); // the hs doesn't answer
		//
		// nokia needs:
		if (state==0) {
		  wlen = writecmd(rd, "AT+CKPD=200\r\n"); // the hs doesn't answer
		  //  wlen = writecmd(rd, "ATA\r"); // the hs doesn't answer
		    //off hook
		} else if (state==1) {
		    wlen = writecmd(rd, "AT+VGM=15\r"); // the hs doesn't answer
		} else if (state==2) {
		    wlen = writecmd(rd, "AT+VGS=15\r"); // the hs doesn't answer
		} else if (state==3) {
		    if ((fd = open(filename,O_RDONLY )) < 0) {
			perror("Can't open input/output file");fflush(stderr);
			fd=-1;
		    }
		} else {
		    wlen = writecmd(rd, "AT+CKPD=200\r"); // the hs doesn't answer
		}
		state++;
	    }

	    if (FD_ISSET(sd, &rfds)) { // master control found
		fprintf(stderr,"sd event\n");
		if (scolink) {
		    fprintf(stderr,"Warning! SCO link already set up!!\n");
		}
		// todo: skip init when set up 
		scolink = sco_socket_accept(sd, &local, &bdaddr, &sco_handle, &sco_mtu);
		if (scolink >0) {
		    maxfd = (maxfd>scolink) ? maxfd : scolink;
		    fprintf(stderr,"Houston, we was connected! %d %d\n",scolink, maxfd);
		} else {
		    fprintf(stderr,"Houston, we have a problem!\n");
		    scolink=0;
		    maxfd = (rd > sd) ? rd : sd;
		}
	    }
	}
    }

    close(sd);
    sleep(1);
    close(rd);
    close(fdout);
    close(fd);
    if (scolink) close(scolink);

    sound_exit();

    return 0;
}
