/*
 *
 *  Computer as Headset utility
 *
 *  Copyright (C) 2009  Stjepan Gros <sgros@zemris.fer.hr>
 *
 *
 *  Based on the example code distributed with PulseAudio.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation;
 *
 *  Software distributed under the License is distributed on an "AS
 *  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 *  implied. See the License for the specific language governing
 *  rights and limitations under the License.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include <pulse/simple.h>
#include <pulse/error.h>
#include <pulse/gccmacro.h>

#define BUFSIZE 1024

/* The Sample format to use */
static const pa_sample_spec ss = {
	.format = PA_SAMPLE_S16LE,
	.rate = 8000,
	.channels = 1
};

static const pa_buffer_attr ba = {
	.maxlength = 0xFFFFFFFF,
	.tlength = 0xFFFFFFFF,
	.prebuf = 0xFFFFFFFF,
	.minreq = 0xFFFFFFFF,
	.fragsize = 0xFFFFFFFF
};

pa_simple *pa_play = NULL;
pa_simple *pa_rec = NULL;

int sound_init(char *appname)
{
	int error;

	/* Create a new playback stream */
	if (!(pa_play = pa_simple_new(NULL, appname, PA_STREAM_PLAYBACK, NULL, "playback", &ss,
			NULL, NULL, &error))) {
		fprintf(stderr, __FILE__": pa_simple_new() failed: %s\n", pa_strerror(error));
		return -1;
	}

	/* Create a new recording stream */
	if (!(pa_rec = pa_simple_new(NULL, appname, PA_STREAM_RECORD, NULL,
			"record", &ss, NULL, NULL, &error))) {
		fprintf(stderr, __FILE__": pa_simple_new() failed: %s\n", pa_strerror(error));
		pa_simple_free(pa_play);
		pa_play = NULL;
		return -1;
	}

	return 0;
}

int sound_play(char *buf, int len)
{
	int error;

	if (pa_simple_write(pa_play, buf, (size_t) len, &error) < 0) {
		fprintf(stderr, __FILE__": pa_simple_write() failed: %s\n", pa_strerror(error));
		return -1;
	}

	return 0;
}

int sound_rec(char *buf, int len)
{
	int error, s;

	/* Record some data ... */
	if ((s = pa_simple_read(pa_rec, buf, (size_t) len, &error)) < 0) {
		fprintf(stderr, __FILE__": pa_simple_read() failed: %s\n", pa_strerror(error));
		return -1;
	}

	return 0;
}

void sound_exit()
{
	int error;

	if (pa_play) {

		/* Make sure that every single sample was played */
		if (pa_simple_drain(pa_play, &error) < 0)
			fprintf(stderr, __FILE__": pa_simple_drain() failed: %s\n", pa_strerror(error));

		pa_simple_free(pa_play);
	}

	if (pa_rec)
		pa_simple_free(pa_rec);
}

void sound_pause(void)
{
}

void sound_resume(void)
{
}

#ifdef DEBUG
int main(int argc, char*argv[])
{
	int ret = 1;
	int error;

	/* replace STDIN with the specified file if needed */
	if (argc > 1) {
		int fd;

		if ((fd = open(argv[1], O_RDONLY)) < 0) {
			fprintf(stderr, __FILE__": open() failed: %s\n", strerror(errno));
			goto finish;
		}

		if (dup2(fd, STDIN_FILENO) < 0) {
			fprintf(stderr, __FILE__": dup2() failed: %s\n", strerror(errno));
			goto finish;
		}

		close(fd);
	}

	if (sound_init(argv[0]) < 0)
		goto finish;

	for (;;) {
		char buf[BUFSIZE];
		ssize_t r;

//		r = sound_rec(buf, 48);
//		fprintf(stderr, "%d\n", r);
//		write(STDOUT_FILENO, buf, 48);
//		continue;

		/* Read some data ... */
		if ((r = read(STDIN_FILENO, buf, sizeof(buf))) <= 0) {
			if (r == 0) /* EOF */
				break;

			fprintf(stderr, __FILE__": read() failed: %s\n", strerror(errno));
			goto finish;
		}

		if (sound_play(buf, r) < 0)
			goto finish;

		printf ("DONE\n");
	}

	/* Make sure that every single sample was played */
	if (pa_simple_drain(pa_play, &error) < 0) {
		fprintf(stderr, __FILE__": pa_simple_drain() failed: %s\n", pa_strerror(error));
		goto finish;
	}

	ret = 0;

finish:
	sound_exit();

	return ret;
}
#endif

